"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.listenWithPgClientFromPgSource = exports.withPgClientFromPgSource = exports.getWithPgClientFromPgSource = void 0;
const grafast_1 = require("grafast");
const isTest = process.env.NODE_ENV === "test";
const pgClientBySourceCache = new Map();
/**
 * Get or build the 'withPgClient' callback function for a given database
 * source, caching it to make future lookups faster.
 */
function getWithPgClientFromPgSource(source) {
    const existing = pgClientBySourceCache.get(source);
    if (existing) {
        if ((0, grafast_1.isPromiseLike)(existing)) {
            return existing.then((v) => {
                v.retainers++;
                return v.withPgClient;
            });
        }
        else {
            existing.retainers++;
            return existing.withPgClient;
        }
    }
    else {
        const promise = (async () => {
            // TODO: We should cache imports
            let adaptor;
            try {
                adaptor = require(source.adaptor);
            }
            catch (e) {
                if (e.code === "ERR_REQUIRE_ESM") {
                    adaptor = await import(source.adaptor);
                }
                else {
                    throw e;
                }
            }
            const factory = adaptor?.createWithPgClient ?? adaptor?.default?.createWithPgClient;
            if (typeof factory !== "function") {
                throw new Error(`'${source.adaptor}' does not look like a withPgClient adaptor - please ensure it exports a method called 'createWithPgClient'`);
            }
            const originalWithPgClient = await factory(source.adaptorSettings);
            const withPgClient = (...args) => originalWithPgClient.apply(null, args);
            const cachedValue = {
                withPgClient,
                retainers: 1,
            };
            let released = false;
            withPgClient.release = () => {
                cachedValue.retainers--;
                // To allow for other promises to resolve and add/remove from the retaininers, check after a tick
                setTimeout(() => {
                    if (cachedValue.retainers === 0 && !released) {
                        released = true;
                        pgClientBySourceCache.delete(source);
                        return originalWithPgClient.release();
                    }
                    // TODO: this used to be zero, but that seems really inefficient...
                    // Figure out why I did that?
                    // }, 0);
                }, isTest ? 500 : 5000);
            };
            pgClientBySourceCache.set(source, cachedValue);
            return cachedValue;
        })();
        if (!pgClientBySourceCache.has(source)) {
            pgClientBySourceCache.set(source, promise);
        }
        promise.catch(() => {
            pgClientBySourceCache.delete(source);
        });
        return promise.then((v) => v.withPgClient);
    }
}
exports.getWithPgClientFromPgSource = getWithPgClientFromPgSource;
async function withPgClientFromPgSource(source, pgSettings, callback) {
    const withPgClientFromPgSource = getWithPgClientFromPgSource(source);
    const withPgClient = (0, grafast_1.isPromiseLike)(withPgClientFromPgSource)
        ? await withPgClientFromPgSource
        : withPgClientFromPgSource;
    try {
        return await withPgClient(pgSettings, callback);
    }
    finally {
        withPgClient.release();
    }
}
exports.withPgClientFromPgSource = withPgClientFromPgSource;
async function listenWithPgClientFromPgSource(source, topic, callback) {
    const deferredUnlisten = (0, grafast_1.defer)();
    withPgClientFromPgSource(source, null, async (client) => {
        if (!client.listen) {
            throw new Error(`Client for '${source.name}' does not support listening`);
        }
        const keepalive = (0, grafast_1.defer)();
        const unlisten = await client.listen(topic, callback);
        deferredUnlisten.resolve(() => {
            unlisten();
            keepalive.resolve();
        });
        return keepalive;
    }).catch((e) => deferredUnlisten.reject(e));
    return deferredUnlisten;
}
exports.listenWithPgClientFromPgSource = listenWithPgClientFromPgSource;
//# sourceMappingURL=pgSources.js.map