"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.PgJWTPlugin = void 0;
require("graphile-config");
const graphile_export_1 = require("graphile-export");
const jsonwebtoken_1 = require("jsonwebtoken");
const behavior_js_1 = require("../behavior.js");
const index_js_1 = require("../index.js");
exports.PgJWTPlugin = {
    name: "PgJWTPlugin",
    description: "Converts a Postgres JWT object type into a GraphQL scalar type containing a signed JWT",
    version: index_js_1.version,
    before: ["PgCodecsPlugin", "PgTablesPlugin"],
    gather: {
        namespace: "pgJWT",
        helpers: {},
        hooks: {
            pgCodecs_PgTypeCodec(info, { pgCodec, pgType }) {
                if (info.options.pgJwtType?.[1] === pgType.typname &&
                    info.options.pgJwtType?.[0] === pgType.getNamespace().nspname) {
                    // It's a JWT type!
                    pgCodec.extensions ||= {};
                    pgCodec.extensions.tags ||= {};
                    pgCodec.extensions.tags.behavior = ["-table", "jwt"];
                }
            },
        },
    },
    schema: {
        hooks: {
            init(_, build) {
                const { options: { pgJwtSecret, pgJwtSignOptions }, } = build;
                const jwtCodec = [...build.pgCodecMetaLookup.keys()].find((codec) => {
                    const behavior = (0, behavior_js_1.getBehavior)(codec.extensions);
                    // TODO: why is b.jwt_token not found here?
                    return build.behavior.matches(behavior, "jwt");
                });
                if (!jwtCodec) {
                    return _;
                }
                const jwtTypeName = build.inflection.tableType(jwtCodec);
                const columnNames = Object.keys(jwtCodec.columns);
                build.registerScalarType(jwtTypeName, {
                    isPgJwtType: true,
                    pgCodec: jwtCodec,
                }, () => ({
                    description: build.wrapDescription("A JSON Web Token defined by [RFC 7519](https://tools.ietf.org/html/rfc7519) which securely represents claims between two parties.", "type"),
                    serialize: (0, graphile_export_1.EXPORTABLE)((columnNames, pgJwtSecret, pgJwtSignOptions, signJwt) => function serialize(value) {
                        const token = columnNames.reduce((memo, columnName) => {
                            if (columnName === "exp") {
                                memo[columnName] = value[columnName]
                                    ? parseFloat(value[columnName])
                                    : undefined;
                            }
                            else {
                                memo[columnName] = value[columnName];
                            }
                            return memo;
                        }, {});
                        const options = Object.assign(Object.create(null), pgJwtSignOptions, token.aud || (pgJwtSignOptions && pgJwtSignOptions.audience)
                            ? null
                            : {
                                audience: "postgraphile",
                            }, token.iss || (pgJwtSignOptions && pgJwtSignOptions.issuer)
                            ? null
                            : {
                                issuer: "postgraphile",
                            }, token.exp ||
                            (pgJwtSignOptions && pgJwtSignOptions.expiresIn)
                            ? null
                            : {
                                expiresIn: "1 day",
                            });
                        return signJwt(token, pgJwtSecret, options);
                    }, [columnNames, pgJwtSecret, pgJwtSignOptions, jsonwebtoken_1.sign]),
                    extensions: {
                        graphile: {
                            // TODO: optimized version of this
                            plan: (0, graphile_export_1.EXPORTABLE)(() => function plan($in) {
                                const $record = $in;
                                return $record.record();
                            }, []),
                        },
                    },
                }), "JWT scalar from PgJWTPlugin");
                build.setGraphQLTypeForPgCodec(jwtCodec, "output", jwtTypeName);
                return _;
            },
        },
    },
};
//# sourceMappingURL=PgJWTPlugin.js.map