"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.PgMutationCreatePlugin = void 0;
require("graphile-config");
const pg_1 = require("@dataplan/pg");
const grafast_1 = require("grafast");
const graphile_export_1 = require("graphile-export");
const behavior_js_1 = require("../behavior.js");
const index_js_1 = require("../index.js");
const utils_js_1 = require("../utils.js");
const isInsertable = (build, source) => {
    if (source.parameters)
        return false;
    if (!source.codec.columns)
        return false;
    if (source.codec.polymorphism)
        return false;
    if (source.codec.isAnonymous)
        return false;
    const behavior = (0, behavior_js_1.getBehavior)(source.extensions);
    return build.behavior.matches(behavior, "insert", "insert") === true;
};
exports.PgMutationCreatePlugin = {
    name: "PgMutationCreatePlugin",
    description: "Adds 'create' mutation for supported table-like sources",
    version: index_js_1.version,
    after: ["smart-tags"],
    inflection: {
        add: {
            createField(options, source) {
                return this.camelCase(`create-${this.tableType(source.codec)}`);
            },
            createInputType(options, source) {
                return this.upperCamelCase(`${this.createField(source)}-input`);
            },
            createPayloadType(options, source) {
                return this.upperCamelCase(`${this.createField(source)}-payload`);
            },
            tableFieldName(options, source) {
                return this.camelCase(`${this.tableType(source.codec)}`);
            },
        },
    },
    schema: {
        hooks: {
            init(_, build) {
                const { inflection, graphql: { GraphQLString, GraphQLNonNull }, } = build;
                const insertableSources = build.input.pgSources.filter((source) => isInsertable(build, source));
                insertableSources.forEach((source) => {
                    build.recoverable(null, () => {
                        const tableTypeName = inflection.tableType(source.codec);
                        const inputTypeName = inflection.createInputType(source);
                        const tableFieldName = inflection.tableFieldName(source);
                        build.registerInputObjectType(inputTypeName, { isMutationInput: true }, () => ({
                            description: `All input for the create \`${tableTypeName}\` mutation.`,
                            fields: ({ fieldWithHooks }) => {
                                const TableInput = build.getGraphQLTypeByPgCodec(source.codec, "input");
                                return {
                                    clientMutationId: {
                                        type: GraphQLString,
                                        applyPlan: (0, graphile_export_1.EXPORTABLE)(() => function plan($input, val) {
                                            $input.set("clientMutationId", val.get());
                                        }, []),
                                    },
                                    ...(TableInput
                                        ? {
                                            [tableFieldName]: fieldWithHooks({
                                                fieldName: tableFieldName,
                                                fieldBehaviorScope: `insert:input:record`,
                                            }, () => ({
                                                description: build.wrapDescription(`The \`${tableTypeName}\` to be created by this mutation.`, "field"),
                                                type: new GraphQLNonNull(TableInput),
                                                applyPlan: (0, graphile_export_1.EXPORTABLE)(() => function plan($object) {
                                                    const $record = $object.getStepForKey("result");
                                                    return $record.setPlan();
                                                }, []),
                                            })),
                                        }
                                        : null),
                                };
                            },
                        }), `PgMutationCreatePlugin input for ${source.name}`);
                        const payloadTypeName = inflection.createPayloadType(source);
                        const behavior = (0, behavior_js_1.getBehavior)(source.extensions);
                        build.registerObjectType(payloadTypeName, {
                            isMutationPayload: true,
                            // TODO: isPgCreatePayloadType: true,
                            pgTypeSource: source,
                        }, grafast_1.ExecutableStep, () => ({
                            description: `The output of our create \`${tableTypeName}\` mutation.`,
                            fields: ({ fieldWithHooks }) => {
                                const TableType = build.getGraphQLTypeByPgCodec(source.codec, "output");
                                return {
                                    clientMutationId: {
                                        type: GraphQLString,
                                        plan: (0, graphile_export_1.EXPORTABLE)((constant) => function plan($mutation) {
                                            return ($mutation.getStepForKey("clientMutationId", true) ?? constant(null));
                                        }, [grafast_1.constant]),
                                    },
                                    ...(TableType &&
                                        build.behavior.matches(behavior, "insert:payload:record", "insert:payload:record")
                                        ? {
                                            [tableFieldName]: fieldWithHooks({
                                                fieldName: tableFieldName,
                                                fieldBehaviorScope: `insert:payload:record`,
                                            }, {
                                                description: `The \`${tableTypeName}\` that was created by this mutation.`,
                                                type: TableType,
                                                plan: (0, graphile_export_1.EXPORTABLE)(() => function plan($object) {
                                                    return $object.get("result");
                                                }, []),
                                                deprecationReason: (0, utils_js_1.tagToString)(source.extensions?.tags?.deprecated),
                                            }),
                                        }
                                        : null),
                                };
                            },
                        }), `PgMutationCreatePlugin payload for ${source.name}`);
                    });
                });
                return _;
            },
            GraphQLObjectType_fields(fields, build, context) {
                const { inflection, graphql: { GraphQLNonNull }, } = build;
                const { scope: { isRootMutation }, fieldWithHooks, } = context;
                if (!isRootMutation) {
                    return fields;
                }
                const insertableSources = build.input.pgSources.filter((source) => isInsertable(build, source));
                return insertableSources.reduce((memo, source) => {
                    return build.recoverable(memo, () => {
                        const createFieldName = inflection.createField(source);
                        const payloadTypeName = inflection.createPayloadType(source);
                        const payloadType = build.getOutputTypeByName(payloadTypeName);
                        const mutationInputType = build.getInputTypeByName(inflection.createInputType(source));
                        return build.extend(memo, {
                            [createFieldName]: fieldWithHooks({ fieldName: createFieldName, fieldBehaviorScope: "insert" }, {
                                args: {
                                    input: {
                                        type: new GraphQLNonNull(mutationInputType),
                                        applyPlan: (0, graphile_export_1.EXPORTABLE)(() => function plan(_, $object) {
                                            return $object;
                                        }, []),
                                    },
                                },
                                type: payloadType,
                                description: `Creates a single \`${inflection.tableType(source.codec)}\`.`,
                                deprecationReason: (0, utils_js_1.tagToString)(source.extensions?.tags?.deprecated),
                                plan: (0, graphile_export_1.EXPORTABLE)((object, pgInsert, source) => function plan(_, args) {
                                    const plan = object({
                                        result: pgInsert(source, {}),
                                    });
                                    args.apply(plan);
                                    return plan;
                                }, [grafast_1.object, pg_1.pgInsert, source]),
                            }),
                        }, `Adding create mutation for ${source.name}`);
                    });
                }, fields);
            },
        },
    },
};
//# sourceMappingURL=PgMutationCreatePlugin.js.map