"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.PgMutationPayloadEdgePlugin = void 0;
require("graphile-build");
require("graphile-config");
const pg_1 = require("@dataplan/pg");
const grafast_1 = require("grafast");
const graphile_export_1 = require("graphile-export");
const behavior_js_1 = require("../behavior.js");
const index_js_1 = require("../index.js");
const utils_js_1 = require("../utils.js");
const PgConnectionArgOrderByPlugin_js_1 = require("./PgConnectionArgOrderByPlugin.js");
exports.PgMutationPayloadEdgePlugin = {
    name: "PgMutationPayloadEdgePlugin",
    description: "Adds 'edge' field to mutation payloads to aid with Relay pagination",
    version: index_js_1.version,
    inflection: {
        add: {
            tableEdgeField(options, codec) {
                return this.camelCase(this.tableEdgeType(codec));
            },
        },
    },
    schema: {
        hooks: {
            GraphQLObjectType_fields(fields, build, context) {
                const { extend, getTypeByName, graphql: { GraphQLList, GraphQLNonNull }, inflection, options: { simpleCollections }, } = build;
                const { scope: { isMutationPayload, pgTypeSource, pgCodec: _pgCodec }, fieldWithHooks, Self, } = context;
                const pgCodec = pgTypeSource?.codec ?? _pgCodec;
                if (!isMutationPayload ||
                    !pgCodec ||
                    !pgCodec.columns ||
                    pgCodec.isAnonymous) {
                    return fields;
                }
                if (pgTypeSource?.parameters && !pgTypeSource.isUnique) {
                    return fields;
                }
                const behavior = (0, behavior_js_1.getBehavior)(pgCodec.extensions);
                const defaultBehavior = simpleCollections === "both"
                    ? "connection list"
                    : simpleCollections === "only"
                        ? "list"
                        : "connection";
                // TODO: this is a rule on the codec; we must ensure that everywhere
                // that uses 'list'/'connection' respects the codec behaviors
                if (!build.behavior.matches(behavior, "*:connection", defaultBehavior)) {
                    return fields;
                }
                const sources = build.input.pgSources.filter((source) => {
                    if (source.codec !== pgCodec)
                        return false;
                    if (source.parameters)
                        return false;
                    return true;
                });
                if (sources.length !== 1) {
                    return fields;
                }
                const source = sources[0];
                const pk = source.uniques?.find((u) => u.isPrimary);
                if (!pk) {
                    return fields;
                }
                const pkColumns = pk.columns;
                const TableType = build.getGraphQLTypeByPgCodec(pgCodec, "output");
                if (!TableType) {
                    return fields;
                }
                const tableTypeName = TableType.name;
                const tableOrderByTypeName = inflection.orderByType(tableTypeName);
                const TableOrderByType = getTypeByName(tableOrderByTypeName);
                if (!TableOrderByType) {
                    return fields;
                }
                const TableEdgeType = getTypeByName(inflection.tableEdgeType(source.codec));
                if (!TableEdgeType) {
                    return fields;
                }
                const fieldName = inflection.tableEdgeField(source.codec);
                const primaryKeyAsc = inflection.builtin("PRIMARY_KEY_ASC");
                const defaultValueEnum = TableOrderByType.getValues().find((v) => v.name === primaryKeyAsc) ||
                    TableOrderByType.getValues()[0];
                return extend(fields, {
                    [fieldName]: fieldWithHooks({
                        fieldName,
                        // TODO: fieldBehaviorScope: `...`,
                        isPgMutationPayloadEdgeField: true,
                        pgCodec: pgCodec,
                    }, () => ({
                        description: build.wrapDescription(`An edge for our \`${tableTypeName}\`. May be used by Relay 1.`, "field"),
                        type: TableEdgeType,
                        args: {
                            orderBy: {
                                description: build.wrapDescription(`The method to use when ordering \`${tableTypeName}\`.`, "arg"),
                                type: new GraphQLNonNull(new GraphQLList(new GraphQLNonNull(TableOrderByType))),
                                defaultValue: defaultValueEnum
                                    ? [defaultValueEnum.value]
                                    : null,
                            },
                        },
                        deprecationReason: (0, utils_js_1.tagToString)(source.extensions?.tags?.deprecated),
                        // TODO: review this plan, it feels overly complex and somewhat hacky.
                        plan: (0, graphile_export_1.EXPORTABLE)((EdgeStep, PgDeleteStep, applyOrderToPlan, connection, constant, pgSelectFromRecord, pkColumns, source, tableOrderByTypeName) => function plan($mutation, args, info) {
                            const $result = $mutation.getStepForKey("result", true);
                            if (!$result) {
                                return constant(null);
                            }
                            const $select = (() => {
                                if ($result instanceof PgDeleteStep) {
                                    return pgSelectFromRecord($result.source, $result.record());
                                }
                                else {
                                    const spec = pkColumns.reduce((memo, columnName) => {
                                        memo[columnName] = $result.get(columnName);
                                        return memo;
                                    }, {});
                                    return source.find(spec);
                                }
                            })();
                            // Perform ordering
                            const $value = args.getRaw("orderBy");
                            applyOrderToPlan($select, $value, info.schema.getType(tableOrderByTypeName));
                            const $connection = connection($select);
                            const $single = $select.single();
                            return new EdgeStep($connection, $single);
                        }, [
                            grafast_1.EdgeStep,
                            pg_1.PgDeleteStep,
                            PgConnectionArgOrderByPlugin_js_1.applyOrderToPlan,
                            grafast_1.connection,
                            grafast_1.constant,
                            pg_1.pgSelectFromRecord,
                            pkColumns,
                            source,
                            tableOrderByTypeName,
                        ]),
                    })),
                }, `Adding edge field for table ${pgCodec.name} to mutation payload '${Self.name}'`);
            },
        },
    },
};
//# sourceMappingURL=PgMutationPayloadEdgePlugin.js.map