"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.PgOrderAllColumnsPlugin = void 0;
require("./PgTablesPlugin.js");
require("graphile-config");
const pg_1 = require("@dataplan/pg");
const graphile_export_1 = require("graphile-export");
const behavior_js_1 = require("../behavior.js");
const index_js_1 = require("../index.js");
// TODO: respect indexes - via behavior?
exports.PgOrderAllColumnsPlugin = {
    name: "PgOrderAllColumnsPlugin",
    description: "Allows ordering by table columns",
    version: index_js_1.version,
    inflection: {
        add: {
            orderByColumnEnum(options, { codec, columnName, variant }) {
                const fieldName = this._columnName({ columnName, codec });
                return this.constantCase(`${fieldName}-${variant}`);
            },
        },
    },
    schema: {
        hooks: {
            GraphQLEnumType_values(values, build, context) {
                const { extend, inflection, options } = build;
                const { scope: { isPgRowSortEnum, pgCodec }, } = context;
                const { orderByNullsLast } = options;
                if (!isPgRowSortEnum ||
                    !pgCodec ||
                    !pgCodec.columns ||
                    pgCodec.isAnonymous) {
                    return values;
                }
                const columns = pgCodec.columns;
                const sources = build.input.pgSources.filter((s) => s.codec === pgCodec && !s.parameters);
                const uniques = sources.flatMap((s) => s.uniques);
                return extend(values, Object.entries(columns).reduce((memo, [columnName, column]) => {
                    const behavior = (0, behavior_js_1.getBehavior)([
                        pgCodec.extensions,
                        column.extensions,
                    ]);
                    // Enable ordering, but don't order by array or range types
                    const defaultBehavior = "orderBy orderBy:* -orderBy:array -orderBy:range";
                    if (!build.behavior.matches(behavior, "attribute:orderBy", defaultBehavior)) {
                        return memo;
                    }
                    if (column.codec.arrayOfCodec) {
                        if (!build.behavior.matches(behavior, "attribute:orderBy:array", defaultBehavior)) {
                            return memo;
                        }
                    }
                    if (column.codec.rangeOfCodec) {
                        if (!build.behavior.matches(behavior, "attribute:orderBy:range", defaultBehavior)) {
                            return memo;
                        }
                    }
                    const isUnique = uniques.some((list) => list.columns[0] === columnName);
                    const ascFieldName = inflection.orderByColumnEnum({
                        column,
                        codec: pgCodec,
                        columnName,
                        variant: "asc",
                    });
                    const descFieldName = inflection.orderByColumnEnum({
                        column,
                        codec: pgCodec,
                        columnName,
                        variant: "desc",
                    });
                    memo = extend(memo, {
                        [ascFieldName]: {
                            extensions: {
                                graphile: {
                                    applyPlan: (0, graphile_export_1.EXPORTABLE)((PgSelectStep, PgUnionAllStep, columnName, isUnique, orderByNullsLast) => (plan) => {
                                        if (!(plan instanceof PgSelectStep) &&
                                            !(plan instanceof PgUnionAllStep)) {
                                            throw new Error("Expected a PgSelectStep or PgUnionAllStep when applying ordering value");
                                        }
                                        plan.orderBy({
                                            attribute: columnName,
                                            direction: "ASC",
                                            ...(orderByNullsLast != null
                                                ? {
                                                    nulls: orderByNullsLast ? "LAST" : "FIRST",
                                                }
                                                : null),
                                        });
                                        if (isUnique) {
                                            plan.setOrderIsUnique();
                                        }
                                    }, [
                                        pg_1.PgSelectStep,
                                        pg_1.PgUnionAllStep,
                                        columnName,
                                        isUnique,
                                        orderByNullsLast,
                                    ]),
                                },
                            },
                        },
                    }, `Adding ascending orderBy enum value for ${pgCodec.name}.`);
                    memo = extend(memo, {
                        [descFieldName]: {
                            extensions: {
                                graphile: {
                                    applyPlan: (0, graphile_export_1.EXPORTABLE)((PgSelectStep, PgUnionAllStep, columnName, isUnique, orderByNullsLast) => (plan) => {
                                        if (!(plan instanceof PgSelectStep) &&
                                            !(plan instanceof PgUnionAllStep)) {
                                            throw new Error("Expected a PgSelectStep or PgUnionAllStep when applying ordering value");
                                        }
                                        plan.orderBy({
                                            attribute: columnName,
                                            direction: "DESC",
                                            ...(orderByNullsLast != null
                                                ? {
                                                    nulls: orderByNullsLast ? "LAST" : "FIRST",
                                                }
                                                : null),
                                        });
                                        if (isUnique) {
                                            plan.setOrderIsUnique();
                                        }
                                    }, [
                                        pg_1.PgSelectStep,
                                        pg_1.PgUnionAllStep,
                                        columnName,
                                        isUnique,
                                        orderByNullsLast,
                                    ]),
                                },
                            },
                        },
                    }, `Adding descending orderBy enum value for ${pgCodec.name}.`);
                    return memo;
                }, {}), `Adding order values from table '${pgCodec.name}'`);
            },
        },
    },
};
//# sourceMappingURL=PgOrderAllColumnsPlugin.js.map