"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.PgRefsPlugin = void 0;
const pg_1 = require("@dataplan/pg");
const grafast_1 = require("grafast");
const index_js_1 = require("../index.js");
const utils_js_1 = require("../utils.js");
exports.PgRefsPlugin = {
    name: "PgRefsPlugin",
    description: "Looks for `@ref` and `@refVia` smart tags and registers the given refs",
    version: index_js_1.version,
    after: ["smart-tags", "PgRelationsPlugin"],
    inflection: {
        add: {
            refSingle(stuff, { refDefinition, identifier }) {
                return (refDefinition.singleRecordFieldName ?? this.singularize(identifier));
            },
            refList(stuff, { refDefinition, identifier }) {
                return (refDefinition.listFieldName ??
                    this.listField(this.pluralize(this.singularize(identifier))));
            },
            refConnection(stuff, { refDefinition, identifier }) {
                return (refDefinition.connectionFieldName ??
                    this.connectionField(this.pluralize(this.singularize(identifier))));
            },
        },
    },
    gather: {
        namespace: "pgRefs",
        helpers: {},
        initialState: () => ({ sourceEvents: [] }),
        hooks: {
            async pgCodecs_PgTypeCodec(info, event) {
                const { pgCodec, pgClass } = event;
                if (!pgClass) {
                    return;
                }
                const { tags } = pgClass.getTagsAndDescription();
                const rawRefs = Array.isArray(tags.ref)
                    ? tags.ref
                    : tags.ref
                        ? [tags.ref]
                        : null;
                if (!rawRefs) {
                    return;
                }
                const refs = rawRefs.map((ref) => (0, utils_js_1.parseSmartTagsOptsString)(ref, 1));
                const extensions = pgCodec.extensions ?? Object.create(null);
                pgCodec.extensions = extensions;
                const refDefinitions = extensions.refDefinitions ?? Object.create(null);
                extensions.refDefinitions = refDefinitions;
                for (const ref of refs) {
                    const { args: [name], params: { to, plural: rawPlural, singular: rawSingular, via: rawVia, behavior, }, } = ref;
                    const singular = rawSingular != null;
                    if (singular && rawPlural != null) {
                        throw new Error(`Both singular and plural were set on ref '${name}'; this isn't valid`);
                    }
                    if (refDefinitions[name]) {
                        throw new Error(`@ref ${name} already registered in ${pgCodec.name}`);
                    }
                    refDefinitions[name] = {
                        singular,
                        graphqlType: to,
                        extensions: {
                            via: rawVia,
                            tags: {
                                behavior,
                            },
                        },
                    };
                }
            },
            async pgTables_PgSource(info, event) {
                info.state.sourceEvents.push(event);
            },
        },
        async main(output, info) {
            for (const event of info.state.sourceEvents) {
                const { databaseName, source, pgClass } = event;
                const getSourceForTableName = async (targetTableIdentifier) => {
                    const nsp = pgClass.getNamespace();
                    if (!nsp) {
                        throw new Error(`Couldn't determine namespace for '${pgClass.relname}'`);
                    }
                    const [targetSchemaName, targetTableName] = (0, utils_js_1.parseDatabaseIdentifierFromSmartTag)(targetTableIdentifier, 2, nsp.nspname);
                    const targetPgClass = await info.helpers.pgIntrospection.getClassByName(databaseName, targetSchemaName, targetTableName);
                    if (!targetPgClass) {
                        return null;
                    }
                    const targetSource = await info.helpers.pgCodecs.getCodecFromClass(databaseName, targetPgClass._id);
                    return targetSource;
                };
                const { tags } = pgClass.getTagsAndDescription();
                const rawRefVias = Array.isArray(tags.refVia)
                    ? tags.refVia
                    : tags.refVia
                        ? [tags.refVia]
                        : null;
                const refDefinitions = source.codec.extensions?.refDefinitions;
                if (!refDefinitions) {
                    if (rawRefVias) {
                        throw new Error(`@refVia without matching @ref is invalid`);
                    }
                    continue;
                }
                const refVias = rawRefVias?.map((refVia) => (0, utils_js_1.parseSmartTagsOptsString)(refVia, 1)) ??
                    [];
                for (const [refName, refDefinition] of Object.entries(refDefinitions)) {
                    const relevantVias = refVias.filter((v) => v.args[0] === refName);
                    const relevantViaStrings = relevantVias
                        .map((v) => v.params.via)
                        .filter((via) => {
                        if (!via) {
                            console.warn(`Invalid '@refVia' detected on '${pgClass.getNamespace().nspname}.${pgClass.relname}'; no 'via:' parameter.`);
                            return false;
                        }
                        else {
                            return true;
                        }
                    });
                    const rawVia = refDefinition.extensions?.via;
                    const vias = [...(rawVia ? [rawVia] : []), ...relevantViaStrings];
                    const paths = [];
                    if (vias.length === 0) {
                        if (!tags.interface) {
                            console.warn(`@ref ${refName} has no valid 'via' on ${source.name}`);
                            continue;
                        }
                    }
                    outerLoop: for (const via of vias) {
                        const path = [];
                        const parts = via.split(";");
                        let currentSource = source;
                        for (const rawPart of parts) {
                            const relationEntries = Object.entries(currentSource.getRelations());
                            const part = rawPart.trim();
                            // TODO: allow whitespace
                            const matches = part.match(/^\(([^)]+)\)->([^)]+)(?:\(([^)]+)\))?$/);
                            let relationEntry;
                            if (matches) {
                                const [, rawLocalCols, targetTableIdentifier, maybeRawTargetCols,] = matches;
                                // TODO: use proper identifier parsing here!
                                const localColumns = rawLocalCols
                                    .split(",")
                                    .map((t) => t.trim());
                                const maybeTargetColumns = maybeRawTargetCols
                                    ? // TODO: use proper identifier parsing here!
                                        maybeRawTargetCols.split(",").map((t) => t.trim())
                                    : null;
                                const targetSource = await getSourceForTableName(targetTableIdentifier);
                                if (!targetSource) {
                                    console.error(`Ref ${refName} has bad via '${via}' which references table '${targetTableIdentifier}' which we either cannot find, or have not generated a source for. Please be sure to indicate the schema if required.`);
                                    continue outerLoop;
                                }
                                relationEntry = relationEntries.find(([, rel]) => {
                                    if (rel.source.name !== targetSource.name) {
                                        return false;
                                    }
                                    if (!(0, grafast_1.arraysMatch)(rel.localColumns, localColumns)) {
                                        return false;
                                    }
                                    if (maybeTargetColumns) {
                                        if (!(0, grafast_1.arraysMatch)(rel.remoteColumns, maybeTargetColumns)) {
                                            return false;
                                        }
                                    }
                                    return true;
                                });
                            }
                            else {
                                const targetTableIdentifier = part;
                                const targetSource = await getSourceForTableName(targetTableIdentifier);
                                if (!targetSource) {
                                    console.error(`Ref ${refName} has bad via '${via}' which references table '${targetTableIdentifier}' which we either cannot find, or have not generated a source for. Please be sure to indicate the schema if required.`);
                                    continue outerLoop;
                                }
                                relationEntry = relationEntries.find(([, rel]) => {
                                    return rel.source.name === targetSource.name;
                                });
                            }
                            if (relationEntry) {
                                path.push({
                                    relationName: relationEntry[0],
                                });
                                const nextSource = relationEntry[1].source;
                                currentSource =
                                    nextSource instanceof pg_1.PgSourceBuilder
                                        ? nextSource.get()
                                        : nextSource;
                            }
                            else {
                                console.warn(`Could not find matching relation for '${via}' / ${currentSource.name} -> '${rawPart}'`);
                                continue outerLoop;
                            }
                        }
                        paths.push(path);
                    }
                    if (source.refs[refName]) {
                        throw new Error(`@ref ${refName} already registered in ${source.name}`);
                    }
                    source.refs[refName] = {
                        definition: refDefinition,
                        paths,
                    };
                }
            }
        },
    },
};
//# sourceMappingURL=PgRefsPlugin.js.map