import "graphile-build";
import type { PgSource, PgSourceOptions, PgSourceRelation, PgSourceUnique, PgTypeCodec, PgTypeColumn } from "@dataplan/pg";
import { PgSourceBuilder } from "@dataplan/pg";
import type { PluginHook } from "graphile-config";
import type { PgClass, PgConstraint, PgNamespace } from "pg-introspection";
declare type PgSourceBuilderOptions = Omit<PgSourceOptions<any, any, any, any>, "relations">;
declare global {
    namespace GraphileBuild {
        interface GraphileBuildSchemaOptions {
            /**
             * If true, setof functions cannot return null, so our list and
             * connection types can use GraphQLNonNull in more places.
             */
            pgForbidSetofFunctionsToReturnNull?: boolean;
        }
        interface BuildInput {
            pgSources: PgSource<any, any, any, any>[];
        }
        interface Inflection {
            /**
             * A PgSource represents a single way of getting a number of values of
             * `source.codec` type. It doesn't necessarily represent a table directly
             * (although it can) - e.g. it might be a function that returns records
             * from a table, or it could be a "sub-selection" of a table, e.g.
             * "admin_users" might be "users where admin is true".  This inflector
             * gives a name to this source, it's primarily used when naming _fields_
             * in the GraphQL schema (as opposed to `_codecName` which typically
             * names _types_.
             *
             * @remarks The method beginning with `_` implies it's not ment to
             * be called directly, instead it's called from other inflectors to give
             * them common behavior.
             */
            _sourceName(this: Inflection, source: PgSource<any, any, any, any>): string;
            /**
             * Takes a `_sourceName` and singularizes it.
             */
            _singularizedSourceName(this: Inflection, source: PgSource<any, any, any, any>): string;
            /**
             * When you're using multiple databases and/or schemas, you may want to
             * prefix various type names/field names with an identifier for these
             * DBs/schemas.
             */
            _schemaPrefix(this: Inflection, details: {
                databaseName: string;
                pgNamespace: PgNamespace;
            }): string;
            /**
             * The name of the PgSource for a table/class
             */
            tableSourceName(this: Inflection, details: {
                databaseName: string;
                pgClass: PgClass;
            }): string;
            /**
             * A PgTypeCodec may represent any of a wide range of PostgreSQL types;
             * this inflector gives a name to this codec, it's primarily used when
             * naming _types_ in the GraphQL schema (as opposed to `_sourceName`
             * which typically names _fields_).
             *
             * @remarks The method beginning with `_` implies it's not ment to
             * be called directly, instead it's called from other inflectors to give
             * them common behavior.
             */
            _codecName(this: Inflection, codec: PgTypeCodec<any, any, any>): string;
            /**
             * Takes a `_codecName` and singularizes it. This is also a good place to
             * try and avoid potential conflicts, e.g. for a table `foo` a `Foo` and
             * `FooInput` and `FooPatch` type might be generated. So a `foo_input`
             * table could potentially cause conflicts. The default inflector would
             * turn `foo_input` into `FooInputRecord`.
             *
             * @remarks The method beginning with `_` implies it's not ment to
             * be called directly, instead it's called from other inflectors to give
             * them common behavior.
             */
            _singularizedCodecName(this: Inflection, codec: PgTypeCodec<any, any, any>): string;
            /**
             * Appends '_record' to a name that ends in `_input`, `_patch`, `Input`
             * or `Patch` to avoid naming conflicts.
             */
            dontEndInInputOrPatch(this: Inflection, text: string): string;
            /**
             * The name of the GraphQL Object Type that's generated to represent a
             * specific table (more specifically a PostgreSQL "pg_class" which is
             * represented as a certain PgTypeCodec)
             */
            tableType(this: GraphileBuild.Inflection, codec: PgTypeCodec<any, any, any>): string;
            tableConnectionType(this: GraphileBuild.Inflection, codec: PgTypeCodec<any, any, any>): string;
            tableEdgeType(this: GraphileBuild.Inflection, codec: PgTypeCodec<any, any, any>): string;
            patchType(this: GraphileBuild.Inflection, typeName: string): string;
            baseInputType(this: GraphileBuild.Inflection, typeName: string): string;
        }
        interface ScopeObject {
            pgCodec?: PgTypeCodec<any, any, any>;
            isPgTableType?: boolean;
            isPgConnectionRelated?: true;
        }
        interface ScopeObjectFieldsField {
            pgSource?: PgSource<any, any, any, any>;
            pgFieldCodec?: PgTypeCodec<any, any, any>;
            pgColumn?: PgTypeColumn<any, any>;
            isPgFieldConnection?: boolean;
            isPgFieldSimpleCollection?: boolean;
        }
        interface ScopeInterfaceFieldsField {
            pgSource?: PgSource<any, any, any, any>;
            pgFieldCodec?: PgTypeCodec<any, any, any>;
            pgColumn?: PgTypeColumn<any, any>;
            isPgFieldConnection?: boolean;
            isPgFieldSimpleCollection?: boolean;
        }
    }
}
declare global {
    namespace GraphileConfig {
        type PgTablesPluginSourceRelations = {
            [identifier: string]: PgSourceRelation<any, any>;
        };
        interface GatherHelpers {
            pgTables: {
                getSourceBuilder(databaseName: string, pgClass: PgClass): Promise<PgSourceBuilder<any, any, any> | null>;
                getSource(sourceBuilder: PgSourceBuilder<any, any, any>): Promise<PgSource<any, any, any, any> | null>;
            };
        }
        interface GatherHooks {
            pgTables_PgSourceBuilder_relations: PluginHook<(event: {
                databaseName: string;
                pgClass: PgClass;
                sourceBuilder: PgSourceBuilder<any, any, any>;
                relations: PgTablesPluginSourceRelations;
            }) => Promise<void> | void>;
            pgTables_PgSource: PluginHook<(event: {
                databaseName: string;
                pgClass: PgClass;
                source: PgSource<any, any, any>;
                relations: PgTablesPluginSourceRelations;
            }) => Promise<void> | void>;
            pgTables_PgSourceBuilder_options: PluginHook<(event: {
                databaseName: string;
                pgClass: PgClass;
                options: PgSourceBuilderOptions;
            }) => void | Promise<void>>;
            pgTables_unique: PluginHook<(event: {
                databaseName: string;
                pgClass: PgClass;
                pgConstraint: PgConstraint;
                unique: PgSourceUnique;
            }) => void | Promise<void>>;
        }
    }
}
export declare const PgTablesPlugin: GraphileConfig.Plugin;
export {};
//# sourceMappingURL=PgTablesPlugin.d.ts.map